#include <winsock2.h>
#include <WS2tcpip.h> // pour inet_pton()
#include <iostream>
#include <string>
using namespace std;

#pragma comment(lib,"Ws2_32.lib")

using socket_t = SOCKET;
bool est_invalide(socket_t sck) {
   return sck == INVALID_SOCKET;
}
using socket_operation_t = int; 
bool est_erreur(socket_operation_t op) {
   return op == SOCKET_ERROR;
}

template <class T>
   void print(ostream &os, T && arg) {
      os << arg;
   }
template <class T, class ... Args>
   void print(ostream &os, T && arg, Args && ... args) {
      print(os, std::forward<T>(arg));
      print(os, std::forward<Args>(args)...);
   }
template <class ... Args>
   auto erreur(Args && ... args) {
      const auto code = WSAGetLastError();
      print(cerr, std::forward<Args>(args)...);
      cerr << "; erreur " << code << endl;
      return code;
   }

void charger_sockets() noexcept {
   const long VERSION_DEMANDEE = MAKEWORD(2,2);
   WSAData configWinsock;
   if (WSAStartup(VERSION_DEMANDEE, &configWinsock) != 0)
      exit(erreur("Demarrage"));
}

void decharger_sockets() noexcept {
   WSACleanup();
}

socket_t creer_socket() {
   auto sck = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   if (est_invalide(sck))
      exit(erreur("Creation"));
   return sck;
}

void fermer_socket(socket_t sck) noexcept {
   shutdown(sck, SD_BOTH);
   closesocket(sck);
}

void connecter
   (socket_t sck, const string &adresse, unsigned short port) {
   sockaddr_in dest = { 0 };
   dest.sin_family = AF_INET;
   inet_pton(AF_INET, adresse.c_str(), &dest.sin_addr.s_addr);
   dest.sin_port = htons(port);
   auto resultat = connect(
      sck, reinterpret_cast<const sockaddr*>(&dest), sizeof(sockaddr_in)
   );
   if (est_erreur(resultat))
      exit(erreur("Connexion a ", adresse, ":", port));
}

string recevoir(socket_t sck, char fin) {
   const int CAPACITE = 64; // arbitraire
   char tampon[CAPACITE];
   string message_recu;
   bool reception_completee = false;
   do {
      auto n = recv(sck, tampon, CAPACITE, 0);
      if (est_erreur(n))
         exit(erreur("Reception"));
      message_recu.append(tampon + 0, tampon + n);
      if (tampon[n-1] == fin)
         reception_completee = true;
   } while(!reception_completee);
   if (!message_recu.empty()) message_recu.pop_back(); // liminer le dlimiteur
   return message_recu;
}

void envoyer(socket_t sck, string msg, char fin) {
   msg += fin;
   while (!msg.empty()) {
      auto plus_recent = send(
         sck, &msg[0], static_cast<int>(msg.size()), 0
      );
      if (est_erreur(plus_recent))
         exit(erreur("Envoi"));
      msg = msg.substr(plus_recent);
   }
}

int main() {
   const string ADRESSE_DU_SERVEUR = "127.0.0.1";
   const unsigned short PORT_DU_SERVEUR = 4321;
   const string MESSAGE = "Ca marche!";
   const char DELIMITEUR_FIN = '#';
   charger_sockets();
   socket_t client = creer_socket();
   connecter(client, ADRESSE_DU_SERVEUR, PORT_DU_SERVEUR);
   envoyer(client, MESSAGE, DELIMITEUR_FIN);
   cout << "Envoi reussi de " << MESSAGE.size() + 1
        << " bytes, delimiteur inclus. Message : \""
        << MESSAGE << '\"' << endl;
   auto recu = recevoir(client, DELIMITEUR_FIN);
   cout << "Reception reussie de " << recu.size() + 1
        << " bytes, delimiteur inclus. Message : \""
        << recu << '\"' << endl;
   fermer_socket(client);
   decharger_sockets();
}
